package fr.jelora.jraspitools.display;

import java.awt.image.BufferedImage;

import fr.jelora.jraspitools.RasPiPin;
import fr.jelora.jraspitools.gpio.GPIODirection;
import fr.jelora.jraspitools.gpio.GPIOException;
import fr.jelora.jraspitools.gpio.GPIOTools;

public class GenericGraphicDisplay {
	
	/*
	 * Afficheur générique 128*64 basé sur le S6B0107 et le S6B0108
	 */
	
	private int width;
	private int height;
	
	private RasPiPin pinRegisterSelect;
	private RasPiPin pinReadWrite;
	private RasPiPin pinEnable;
	private RasPiPin pinData0;
	private RasPiPin pinData1;
	private RasPiPin pinData2;
	private RasPiPin pinData3;
	private RasPiPin pinData4;
	private RasPiPin pinData5;
	private RasPiPin pinData6;
	private RasPiPin pinData7;
	private RasPiPin pinChipSelect1;
	private RasPiPin pinChipSelect2;
	private RasPiPin pinReset;
	
	private GPIODirection directionData0;
	private GPIODirection directionData1;
	private GPIODirection directionData2;
	private GPIODirection directionData3;
	private GPIODirection directionData4;
	private GPIODirection directionData5;
	private GPIODirection directionData6;
	private GPIODirection directionData7;
	
	private GPIOTools gpioTools;
	
	public GenericGraphicDisplay(int width, int height, RasPiPin pinRegisterSelect, RasPiPin pinReadWrite, RasPiPin pinEnable,
			RasPiPin pinData0, RasPiPin pinData1, RasPiPin pinData2, RasPiPin pinData3, 
			RasPiPin pinData4, RasPiPin pinData5, RasPiPin pinData6, RasPiPin pinData7,
			RasPiPin pinChipSelect1, RasPiPin pinChipSelect2, RasPiPin pinReset) throws DisplayException{
		this.init(width, height, pinRegisterSelect, pinReadWrite, pinEnable, pinData0, pinData1, pinData2, pinData3, pinData4, pinData5, pinData6, pinData7, pinChipSelect1, pinChipSelect2, pinReset);
	}
	
	private void init(int width, int height, RasPiPin pinRegisterSelect, RasPiPin pinReadWrite, RasPiPin pinEnable,
			RasPiPin pinData0, RasPiPin pinData1, RasPiPin pinData2, RasPiPin pinData3, 
			RasPiPin pinData4, RasPiPin pinData5, RasPiPin pinData6, RasPiPin pinData7,
			RasPiPin pinChipSelect1, RasPiPin pinChipSelect2, RasPiPin pinReset) throws DisplayException {
		gpioTools = new GPIOTools();
		
		this.width = width;
		if ( width < 1 ){
			throw new DisplayException("wrong width");
		}
		
		this.height = height;
		if ( height < 1 ){
			throw new DisplayException("wrong height");
		}
		
		try{
			this.pinRegisterSelect = pinRegisterSelect;
			if ( this.pinRegisterSelect != null ){
				gpioTools.init(this.pinRegisterSelect, GPIODirection.out, false, false);
			}
			else{
				throw new DisplayException("pinRegisterSelect not selected");
			}
			
			this.pinReadWrite = pinReadWrite;
			if ( this.pinReadWrite != null ){
				gpioTools.init(this.pinReadWrite, GPIODirection.out, false, false);
			}
			else{
				throw new DisplayException("pinReadWrite not selected");
			}
			
			this.pinEnable = pinEnable;
			if ( this.pinEnable != null ){
				gpioTools.init(this.pinEnable, GPIODirection.out, false, false);
			}
			else{
				throw new DisplayException("pinEnable not selected");
			}
			
			this.pinData0 = pinData0;
			if ( this.pinData0 != null ){
				directionData0 = GPIODirection.out;
				gpioTools.init(this.pinData0, directionData0, false, false);
			}
			else{
				throw new DisplayException("pinData0 not selected");
			}
			
			this.pinData1 = pinData1;
			if ( this.pinData1 != null ){
				directionData1 = GPIODirection.out;
				gpioTools.init(this.pinData1, directionData1, false, false);
			}
			else{
				throw new DisplayException("pinData1 not selected");
			}
			
			this.pinData2 = pinData2;
			if ( this.pinData2 != null ){
				directionData2 = GPIODirection.out;
				gpioTools.init(this.pinData2, directionData2, false, false);
			}
			else{
				throw new DisplayException("pinData2 not selected");
			}
			
			this.pinData3 = pinData3;
			if ( this.pinData3 != null ){
				directionData3 = GPIODirection.out;
				gpioTools.init(this.pinData3, directionData3, false, false);
			}
			else{
				throw new DisplayException("pinData3 not selected");
			}
			
			this.pinData4 = pinData4;
			if ( this.pinData4 != null ){
				directionData4 = GPIODirection.out;
				gpioTools.init(this.pinData4, directionData4, false, false);
			}
			else{
				throw new DisplayException("pinData4 not selected");
			}
			
			this.pinData5 = pinData5;
			if ( this.pinData5 != null ){
				directionData5 = GPIODirection.out;
				gpioTools.init(this.pinData5, directionData5, false, false);
			}
			else{
				throw new DisplayException("pinData5 not selected");
			}
			
			this.pinData6 = pinData6;
			if ( this.pinData6 != null ){
				directionData6 = GPIODirection.out;
				gpioTools.init(this.pinData6, directionData6, false, false);
			}
			else{
				throw new DisplayException("pinData6 not selected");
			}
			
			this.pinData7 = pinData7;
			if ( this.pinData7 != null ){
				directionData7 = GPIODirection.out;
				gpioTools.init(this.pinData7, directionData7, false, false);
			}
			else{
				throw new DisplayException("pinData7 not selected");
			}
			
			this.pinChipSelect1 = pinChipSelect1;
			if ( this.pinChipSelect1 != null ){
				gpioTools.init(this.pinChipSelect1, GPIODirection.out, false, false);
			}
			else{
				throw new DisplayException("pinChipSelect1 not selected");
			}
			
			this.pinChipSelect2 = pinChipSelect2;
			if ( this.pinChipSelect2 != null ){
				gpioTools.init(this.pinChipSelect2, GPIODirection.out, false, false);
			}
			else{
				throw new DisplayException("pinChipSelect2 not selected");
			}
			
			this.pinReset = pinReset;
			if ( this.pinReset != null ){
				gpioTools.init(this.pinReset, GPIODirection.out, false, true);
			}
			else{
				throw new DisplayException("pinReset not selected");
			}
		}
		catch (GPIOException e) {
			throw new DisplayException("error during init GPIOs", e);
		}
		
		// Initialisation
		this.reset();
		this.displayOnOff(true);
	}
	
	private void pauseInstructionCycle() throws DisplayException{
		try {
			Thread.sleep(0, 1000);
		}
		catch (InterruptedException e) {
			throw new DisplayException("error during waiting instruction cycle", e);
		}
	}
	
	public void sendPulseEnable() throws DisplayException{
		try{
			gpioTools.setValue(pinEnable, true);
			pauseInstructionCycle();
			gpioTools.setValue(pinEnable, false);
			pauseInstructionCycle();
		}
		catch(GPIOException e){
			throw new DisplayException("error using GPIO during sending pulse to enable", e);
		}
	}
	
	private void setDirectionDataBus(GPIODirection direction) throws GPIOException{
		if ( directionData0 != direction ){
			gpioTools.setDirection(pinData0, direction);
		}
		if ( directionData1 != direction ){
			gpioTools.setDirection(pinData1, direction);
		}
		if ( directionData2 != direction ){
			gpioTools.setDirection(pinData2, direction);
		}
		if ( directionData3 != direction ){
			gpioTools.setDirection(pinData3, direction);
		}
		if ( directionData4 != direction ){
			gpioTools.setDirection(pinData4, direction);
		}
		if ( directionData5 != direction ){
			gpioTools.setDirection(pinData5, direction);
		}
		if ( directionData6 != direction ){
			gpioTools.setDirection(pinData6, direction);
		}
		if ( directionData7 != direction ){
			gpioTools.setDirection(pinData7, direction);
		}
	}
	
	private void sendByteToDataBus8bits(byte code) throws DisplayException{
		try{
			setDirectionDataBus(GPIODirection.out);
			
			gpioTools.setValue(pinData0, (code & 1) > 0 );
			gpioTools.setValue(pinData1, (code & 2) > 0 );
			gpioTools.setValue(pinData2, (code & 4) > 0 );
			gpioTools.setValue(pinData3, (code & 8) > 0 );
			gpioTools.setValue(pinData4, (code & 16) > 0 );
			gpioTools.setValue(pinData5, (code & 32) > 0 );
			gpioTools.setValue(pinData6, (code & 64) > 0 );
			gpioTools.setValue(pinData7, (code & 128) > 0 );
		}
		catch(GPIOException e){
			throw new DisplayException("error using GPIO during sending data to display databus", e);
		}
	}
	
	public void sendInstruction(byte code, DisplayChipSelect chipSelect) throws DisplayException{
		try{
			gpioTools.setValue(pinRegisterSelect, false); // Registre instruction
			gpioTools.setValue(pinReadWrite, false); // En écriture
			
			if ( chipSelect == DisplayChipSelect.CS1 ){
				gpioTools.setValue(pinChipSelect1, true);
				gpioTools.setValue(pinChipSelect2, false);
			}
			else{
				gpioTools.setValue(pinChipSelect1, false);
				gpioTools.setValue(pinChipSelect2, true);
			}
			
			sendByteToDataBus8bits(code);
			sendPulseEnable();
			
			if ( chipSelect == DisplayChipSelect.CS1 ){
				gpioTools.setValue(pinChipSelect1, false);
			}
			else{
				gpioTools.setValue(pinChipSelect2, false);
			}
		}
		catch(GPIOException e){
			throw new DisplayException("error during sending instruction", e);
		}
	}
	
	public void sendData(byte code, DisplayChipSelect chipSelect) throws DisplayException{
		try{
			gpioTools.setValue(pinRegisterSelect, true); // Registre data
			gpioTools.setValue(pinReadWrite, false); // En écriture
			
			if ( chipSelect == DisplayChipSelect.CS1 ){
				gpioTools.setValue(pinChipSelect1, true);
				gpioTools.setValue(pinChipSelect2, false);
			}
			else{
				gpioTools.setValue(pinChipSelect1, false);
				gpioTools.setValue(pinChipSelect2, true);
			}
			
			sendByteToDataBus8bits(code);
			sendPulseEnable();
			
			if ( chipSelect == DisplayChipSelect.CS1 ){
				gpioTools.setValue(pinChipSelect1, false);
			}
			else{
				gpioTools.setValue(pinChipSelect2, false);
			}
		}
		catch(GPIOException e){
			throw new DisplayException("error during sending instruction", e);
		}
	}
	
	public byte readData(DisplayChipSelect chipSelect) throws DisplayException{
		int value = 0;
		try{
			gpioTools.setValue(pinRegisterSelect, true); // Registre data
			gpioTools.setValue(pinReadWrite, true); // En lecture
			
			if ( chipSelect == DisplayChipSelect.CS1 ){
				gpioTools.setValue(pinChipSelect1, true);
				gpioTools.setValue(pinChipSelect2, false);
			}
			else{
				gpioTools.setValue(pinChipSelect1, false);
				gpioTools.setValue(pinChipSelect2, true);
			}
			
			setDirectionDataBus(GPIODirection.in);
			
			gpioTools.setValue(pinEnable, true);
			pauseInstructionCycle();
			
			value = gpioTools.getValueInteger(pinData0);
			value += gpioTools.getValueInteger(pinData1) * 2;
			value += gpioTools.getValueInteger(pinData2) * 4;
			value += gpioTools.getValueInteger(pinData3) * 8;
			value += gpioTools.getValueInteger(pinData4) * 16;
			value += gpioTools.getValueInteger(pinData5) * 32;
			value += gpioTools.getValueInteger(pinData6) * 64;
			value += gpioTools.getValueInteger(pinData7) * 128;
			
			gpioTools.setValue(pinEnable, false);
			pauseInstructionCycle();
			
			
			if ( chipSelect == DisplayChipSelect.CS1 ){
				gpioTools.setValue(pinChipSelect1, false);
			}
			else{
				gpioTools.setValue(pinChipSelect2, false);
			}
		}
		catch(GPIOException e){
			throw new DisplayException("error during sending instruction", e);
		}
		return (byte) value;
	}
	
	public void reset() throws DisplayException{
		try{
			gpioTools.setValue(pinReset, false);
			sendPulseEnable();
			gpioTools.setValue(pinReset, true);
			sendPulseEnable();
		}
		catch(GPIOException e){
			throw new DisplayException("error during reset", e);
		}
	}
	
	public void displayOnOff(boolean on) throws DisplayException{
		if ( on ){
			this.sendInstruction((byte) 0x3F, DisplayChipSelect.CS1);
			this.sendInstruction((byte) 0x3F, DisplayChipSelect.CS2);
		}
		else{
			this.sendInstruction((byte) 0x3E, DisplayChipSelect.CS1);
			this.sendInstruction((byte) 0x3E, DisplayChipSelect.CS2);
		}
	}
	
	public void setAddress(byte adress, DisplayChipSelect chipSelect) throws DisplayException{
		this.sendInstruction((byte) (0x40 + adress), chipSelect);
	}
	
	public void setPage(byte page, DisplayChipSelect chipSelect) throws DisplayException{
		this.sendInstruction((byte) (0xB8 + page), chipSelect);
	}
	
	public void setDisplayStartLine(byte line, DisplayChipSelect chipSelect) throws DisplayException{
		this.sendInstruction((byte) (0xC0 + line), chipSelect);
	}
	
	public void clear() throws DisplayException{
		for ( int i = 0 ; i <= 7 ; i++ ){
			this.setPage((byte) i, DisplayChipSelect.CS1);
			this.setPage((byte) i, DisplayChipSelect.CS2);
			this.setAddress((byte) 0x00, DisplayChipSelect.CS1);
			this.setAddress((byte) 0x00, DisplayChipSelect.CS2);
			for ( int j = 0 ; j <= 63 ; j++ ){
				this.sendData((byte) 0x00, DisplayChipSelect.CS1);
			}
			for ( int j = 0 ; j <= 63 ; j++ ){
				this.sendData((byte) 0x00, DisplayChipSelect.CS2);
			}
		}
	}
	
	public void drawImage(BufferedImage image) throws DisplayException{
		this.drawImage(image, 0, 0);
	}
	
	public void drawImage(BufferedImage image, int x, int y) throws DisplayException{
		if ( x < 0 || x > width ){
			throw new DisplayException("bad x coordinate");
		}
		if ( y < 0 || y > height ){
			throw new DisplayException("bad y coordinate");
		}
		
		int lignePage = 0;
		int xLigne = 0;
		int xg = 0;
		int yg = 0;
		DisplayChipSelect chipSelect = null;
		byte valeur = 0;
		for ( int page = y / 8 ; page <= 7 && page*8 <= image.getHeight() ; page++ ){
			
			xg = 0;
			for ( xLigne = x ; xLigne <= 127 && xg < image.getWidth() ; xLigne++ ){
				if ( xLigne < 64 ){
					chipSelect = DisplayChipSelect.CS1;
					this.setAddress((byte) xLigne, chipSelect);
				}
				else{
					chipSelect = DisplayChipSelect.CS2;
					this.setAddress((byte) (xLigne - 64), chipSelect);
				}
				this.setPage((byte) page, chipSelect);
				
				valeur = 0;
				for ( lignePage = y - page*8 ; lignePage <= 7 ; lignePage++ ){
					yg = page*8 + lignePage - y;
					if ( yg < image.getHeight() ){
						if ( image.getRGB(xg, yg) == -1 ){
							valeur += (byte) Math.pow(2, lignePage);
						}
					}
				}
				this.sendData(valeur, chipSelect);
				
				xg++;
			}
		}
	}
}
